package cz.mendelu.dnaAnalyser.sequence.annotation;

import cz.mendelu.dnaAnalyser.batch.BatchService;
import cz.mendelu.dnaAnalyser.sequence.NCBISequenceService;
import cz.mendelu.dnaAnalyser.sequence.annotation.feature.Feature;
import cz.mendelu.dnaAnalyser.sequence.annotation.feature.FeatureDataSessionService;
import cz.mendelu.dnaAnalyser.sequence.annotation.feature.FeatureMapper;
import cz.mendelu.dnaAnalyser.utils.exception.NotFoundException;
import cz.mendelu.dnaAnalyser.utils.repository.BaseRepository;
import cz.mendelu.dnaAnalyser.utils.service.ModelService;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.session.SqlSession;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Service
@Slf4j
public class AnnotationService extends ModelService<Annotation> {
    @Autowired
    private NCBISequenceService ncbiSequenceService;

    @Autowired
    private AnnotationRepository annotationRepository;

    @Autowired
    private FeatureDataSessionService featureDataSessionService;

    @Autowired
    private BatchService batchService;

    @Override
    protected BaseRepository<Annotation, UUID> getRepository() {
        return annotationRepository;
    }

    /**
     * Based on provided annotation sequence containing NCBI id downloads and processes (save do DB) annotation
     *
     * @param annotation with sequence with NCBI id.
     * @return processed annotation .
     */
    public Annotation executeGetAnnotation(Annotation annotation) {
          if(annotation.getSequence().getNcbi() == null){
            throw new IllegalArgumentException("Sequence does not contain NCBI id.");
        }
        annotation = annotationRepository.save(annotation);
        Annotation finalAnnotation = annotation;
        batchService.<Annotation>createBatch(annotation.getOwner(), annotation)
                .execute(uuid -> {
                    Annotation processedAnnotation = annotationRepository.findById(uuid)
                            .orElseThrow(() -> new NotFoundException(Annotation.class, uuid));
                    try (SqlSession sqlSession = featureDataSessionService.openDataSession(processedAnnotation)) {

                        processedAnnotation.setType(finalAnnotation.getType());
                        processedAnnotation.setStart(finalAnnotation.getStart());
                        processedAnnotation.setStop(finalAnnotation.getStop());
                        parseAnnotation(finalAnnotation.getSequence().getNcbi(), finalAnnotation.getType(), finalAnnotation.getStart(), finalAnnotation.getStop(), processedAnnotation, sqlSession);

                        sqlSession.commit();
                    }
                    annotationRepository.save(processedAnnotation);
                }).start();

        return annotation;
    }
    /**
     * Parse exact annotation and its features
     *
     * @param id NCBI id for download by ncbiSequenceService.
     * @param type in case of looking only for certain feature type in annotation.
     * @param start lower boundary for specification of desired position.
     * @param stop upper boundary for specification of desired position.
     * @param annotation to which features will be assigned.
     * @param sqlSession opened feature session to handle saving newly created features.

     * @return list of newly created features.
     */
    // Parse exact annotation and its features
    public List<Map<String, Object>> parseAnnotation(String id, String type, Integer start, Integer stop, Annotation annotation, SqlSession sqlSession){
        FeatureMapper featureMapper = sqlSession.getMapper(FeatureMapper.class);

        List<Map<String, Object>> resultList = new ArrayList<>();
        // Accesses NCBI database and downloads by received ID
        Stream<String> result =  ncbiSequenceService.downloadFeature(id);
        List<String> list = result
                .map (String::new)
                .collect(Collectors.toList());
        String featureType = "";

        // Creates list of separated different features contained in annotation
        for(String row : list){
            String[] featureRow = row.split("\t");
            List<String> featureRowList = new ArrayList<>(Arrays.asList(featureRow));
            if(featureRow.length == 3){
                featureType = featureRowList.get(2);
                parseRow(type, featureType, featureRowList, resultList, start, stop, annotation, featureMapper);
            }
            else if (featureRow.length == 2){
                parseRow(type, featureType, featureRowList, resultList, start, stop, annotation, featureMapper);
            }
            sqlSession.commit();
        }
        return resultList;
    }

    /**
     *  Parses one row of downloaded annotation and creates feature from result which is then saved into DB.
     *
     * @param type in case of looking only for certain feature type in annotation.
     * @param featureType retrieved from downloaded annotation used for comparison in case of looking for certain feature type.
     * @param featureRowList one row of an annotation splitted by /tab.
     * @param resultList collection of previous method updated by inserting new features.
     * @param start lower boundary for specification of desired position.
     * @param stop upper boundary for specification of desired position.
     * @param annotation to which features will be assigned.
     * @param featureMapper mapper class for features.
     */
    private void parseRow(String type, String featureType, List<String> featureRowList, List<Map<String, Object>> resultList, Integer start, Integer stop, Annotation annotation, FeatureMapper featureMapper){
        Map<String, Object> featureMap = new HashMap<>();
        if(type == null || featureType.equals(type)) {
            featureRowList.add(featureType);
            //Removes unwanted characters or swapped positions
            List<String> parsedRow = parseAnnotationRow(featureRowList);
            if (checkStartStopCondition(parsedRow, start ,stop)){
                featureMap.put("type", featureType);
                featureMap.put("positions", (new String[]{parsedRow.get(0), parsedRow.get(1)}));
                resultList.add(featureMap);
                Feature feature = new Feature(null, featureType, parsedRow.get(0), parsedRow.get(1), annotation.getId().toString());

                featureMapper.insert(feature);
            }
        }
    }

    /**
     * Processes list of features for pageable response based on received parameter
     *
     * @param list of features from DB.
     * @param pageable pram for configuring response pages.
     * @return page based on parameter containing features adjusted in map.
     */
    public Page<Map<String, Object>> retrieveFeatures(List<Feature> list, Pageable pageable){
        List<Map<String, Object>> resultList = new ArrayList<>();

        list.forEach(feature -> {
            Map<String, Object> featureMap = new HashMap<>();
            featureMap.put("type", feature.getFeatureType());
            featureMap.put("positions", (new String[]{feature.getStartPosition(), feature.getEndPosition()}));
            resultList.add(featureMap);
        });

        int start = (int)pageable.getOffset();
        int end = Math.min((start + pageable.getPageSize()), resultList.size());

        return new PageImpl<>(resultList.subList(start, end), pageable, resultList.size());
    }

    private List<String> parseAnnotationRow(List<String> row){
        int start = Integer.parseInt(row.get(0).replace(">", "").replace("<", "").replace("\n", ""));
        int stop = Integer.parseInt(row.get(1).replace(">", "").replace("<", "").replace("\n", ""));
        String feature = row.get(2).replace("\n", "");

        if (start < stop){
            return new ArrayList<>(Arrays.asList(Integer.toString(start), Integer.toString(stop), feature));
        }else{
            return new ArrayList<>(Arrays.asList(Integer.toString(stop), Integer.toString(start), feature));
        }
    }

    private boolean checkStartStopCondition(List<String> row, Integer start, Integer stop){
        if (start != null & stop != null){
            return Integer.parseInt(row.get(0)) > start && Integer.parseInt(row.get(1))  < stop;
        }
        else if (start != null){
            return Integer.parseInt(row.get(0)) > start;
        }
        else if (stop != null){
            return Integer.parseInt(row.get(1)) < stop;
        }
        else {
            return true;
        }
    }
    /**
     * Get features belonging to exact annotation.
     *
     * @param annotation by which features are to be obtained from DB.
     * @return list of features
     */
    public List<Feature> getFeatures(Annotation annotation) {

        try (SqlSession sqlSession = featureDataSessionService.openDataSession(annotation)) {
            FeatureMapper featureMapper = sqlSession.getMapper(FeatureMapper.class);
            List<Feature> list = featureMapper.getByAnnotation(annotation.getId().toString());
            return list;
        }
    }
}
