package cz.mendelu.dnaAnalyser.user;

import cz.mendelu.dnaAnalyser.utils.controller.*;
import cz.mendelu.dnaAnalyser.jwt.JwtTokenUtil;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.Data;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpStatus;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.bind.annotation.RestController;

import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import java.util.UUID;

import static cz.mendelu.dnaAnalyser.utils.controller.Response.response;

@RestController
@Api(tags = "api/user")
@RequestMapping("/api")
public class UserController {

    @Data
    static class Register {

        @NotNull
        private String login;

        @NotNull
        private String password;

    }

    @Data
    static class ChangePassword {

        @NotNull
        private String oldPassword;

        @NotNull
        private String newPassword;

    }

    @Data
    static class ResetPassword {

        @NotNull
        private String login;

    }

    @Value("${security.jwt.header}")
    private String tokenHeader;

    @Autowired
    private JwtTokenUtil jwtTokenUtil;

    @Autowired
    private UserService userService;

    @PostMapping("/user")
    @ResponseStatus(HttpStatus.CREATED)
    @ApiOperation(value = "Register new user")
    public Response<User> postRegistrationNewUser(
            @Valid @RequestBody Register request
    ) {
        User newUser = userService.registrationUser(request.getLogin(), request.getPassword());
        return response(newUser);
    }

    @GetMapping("/user")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get current user by JWT token")
    public Response<User> getCurrentUser(
            HttpServletRequest request
    ) {
        String token = request.getHeader(tokenHeader);
        UUID userId = jwtTokenUtil.getUserId(token);
        User user = userService.findOne(userId);
        return response(user);
    }

    @DeleteMapping("/user")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("hasRole('ROLE_USER')")
    @ApiOperation("Delete user, current user is defined by JWT token")
    public void deleteUser(
            HttpServletRequest request
    ) {
        String token = request.getHeader(tokenHeader);
        UUID id = jwtTokenUtil.getUserId(token);
        userService.delete(id);
    }

    @PutMapping("/user/change-password")
    @ResponseStatus(HttpStatus.ACCEPTED)
    @PreAuthorize("hasRole('ROLE_USER')")
    @ApiOperation("Change user password for current user by JWT token")
    public Response<User> putChangeUserPassword(
            HttpServletRequest request,
            @Valid @RequestBody ChangePassword requestBody
    ) {
        String token = request.getHeader(tokenHeader);
        UUID userId = jwtTokenUtil.getUserId(token);
        User user = userService.changePassword(userId, requestBody.getOldPassword(), requestBody.getNewPassword());
        return response(user);
    }

    @PutMapping("/user/reset-password")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @ApiOperation("Reset user password.")
    public void putResetUserPassword(
            @Valid @RequestBody ResetPassword request
    ) {
        userService.resetPassword(request.getLogin());
    }
}
